import {
  CharacterEdgeVoiceActorRolesArgs,
  CharacterEdgeVoiceActorsArgs,
  CharacterMediaArgs,
} from "../types/generated/graphql.ts"
import { Fields, Fn, UpdateOperation } from "../types/Anilist.ts"
import { FuzzyDate, PageInfo } from "../anilist.ts"
import { Media, MediaConnection } from "./Media.ts"
import { Staff, StaffRoleType } from "./Staff.ts"

const CharacterName = ({ query, level }: Fields<UpdateOperation>) => ({
  /** The person's given name */
  withFirst() {
    query[0] = query[0].set({ subField: "first", level })
    return this
  },
  /** The person's middle name */
  withMiddle() {
    query[0] = query[0].set({ subField: "middle", level })
    return this
  },
  /** The person's surname */
  withLast() {
    query[0] = query[0].set({ subField: "last", level })
    return this
  },
  /** The person's first and last name */
  withFull() {
    query[0] = query[0].set({ subField: "full", level })
    return this
  },
  /** The person's full name in their native language */
  withNative() {
    query[0] = query[0].set({ subField: "native", level })
    return this
  },
  /** Other names the staff member might be referred to as (pen names) */
  withAlternative() {
    query[0] = query[0].set({ subField: "alternative", level })
    return this
  },
  /** The currently authenticated users preferred name language. Default romaji for non-authenticated */
  withUserPreferred() {
    query[0] = query[0].set({ subField: "userPreferred", level })
    return this
  },
})

const CharacterImage = ({ query, level }: Fields<UpdateOperation>) => ({
  /** The person's image of media at its largest size */
  withLarge() {
    query[0] = query[0].set({ subField: "large", level })
    return this
  },
  /** The person's image of media at medium size */
  withMedium() {
    query[0] = query[0].set({ subField: "medium", level })
    return this
  },
})

const CharacterEdge = ({ query, level }: Fields<UpdateOperation>) => ({
  withNode(fn: Fn<typeof Character>) {
    query[0] = query[0].set({ subField: "node", level, hasSubField: true })
    let tmpQuery
    fn(Character({ query: tmpQuery = [query[0]], level: level + 1 }))
    query[0] = tmpQuery[0]
    return this
  },
  /** The id of the connection */
  withId() {
    query[0] = query[0].set({ subField: "id", level })
    return this
  },
  /** The characters role in the media */
  withRole() {
    query[0] = query[0].set({ subField: "role", level })
    return this
  },
  /** Media specific character name */
  withName() {
    query[0] = query[0].set({ subField: "name", level })
    return this
  },
  /** The voice actors of the character */
  withVoiceActors(args: CharacterEdgeVoiceActorsArgs | undefined, fn: Fn<typeof Staff>) {
    query[0] = query[0].set({ subField: "voiceActors", level, variables: args, hasSubField: true })
    let tmpQuery
    fn(Staff({ query: tmpQuery = [query[0]], level: level + 1 }))
    query[0] = tmpQuery[0]
    return this
  },
  /** The voice actors of the character with role date */
  withVoiceActorRoles(
    args: CharacterEdgeVoiceActorRolesArgs | undefined,
    fn: Fn<typeof StaffRoleType>,
  ) {
    query[0] = query[0].set({
      subField: "voiceActorRoles",
      level,
      variables: args,
      hasSubField: true,
    })
    let tmpQuery
    fn(StaffRoleType({ query: tmpQuery = [query[0]], level: level + 1 }))
    query[0] = tmpQuery[0]
    return this
  },
  /** The media the character is in */
  withMedia(fn: Fn<typeof Media>) {
    query[0] = query[0].set({ subField: "media", level, hasSubField: true })
    let tmpQuery
    fn(Media({ query: tmpQuery = [query[0]], level: level + 1 }))
    query[0] = tmpQuery[0]
    return this
  },
  /** The order the character should be displayed from the users favourites */
  withFavouriteOrder() {
    query[0] = query[0].set({ subField: "favouriteOrder", level })
    return this
  },
})

export const CharacterConnection = ({ query, level }: Fields<UpdateOperation>) => ({
  withEdges(fn: Fn<typeof CharacterEdge>) {
    query[0] = query[0].set({ subField: "edges", level, hasSubField: true })
    let tmpQuery
    fn(CharacterEdge({ query: tmpQuery = [query[0]], level: level + 1 }))
    query[0] = tmpQuery[0]
    return this
  },
  withNodes(fn: Fn<typeof Character>) {
    query[0] = query[0].set({ subField: "nodes", level, hasSubField: true })
    let tmpQuery
    fn(Character({ query: tmpQuery = [query[0]], level: level + 1 }))
    query[0] = tmpQuery[0]
    return this
  },
  /** The pagination information */
  withPageInfo(fn: Fn<typeof PageInfo>) {
    query[0] = query[0].set({ subField: "pageInfo", level, hasSubField: true })
    let tmpQuery
    fn(PageInfo({ query: tmpQuery = [query[0]], level: level + 1 }))
    query[0] = tmpQuery[0]
    return this
  },
})

export const Character = ({ query, level }: Fields<UpdateOperation>) => ({
  /** The id of the character */
  withId() {
    query[0] = query[0].set({ subField: "id", level })
    return this
  },
  /** The names of the character */
  withName(fn: Fn<typeof CharacterName>) {
    query[0] = query[0].set({ subField: "name", level, hasSubField: true })
    let tmpQuery
    fn(CharacterName({ query: tmpQuery = [query[0]], level: level + 1 }))
    query[0] = tmpQuery[0]
    return this
  },
  /** Character images */
  withImage(fn: Fn<typeof CharacterImage>) {
    query[0] = query[0].set({ subField: "image", level, hasSubField: true })
    let tmpQuery
    fn(CharacterImage({ query: tmpQuery = [query[0]], level: level + 1 }))
    query[0] = tmpQuery[0]
    return this
  },
  /** A general description of the character */
  withDescription(args?: { asHtml: boolean }) {
    query[0] = query[0].set({ subField: "description", level, variables: args })
    return this
  },
  /** The character's gender. Usually Male, Female, or Non-binary but can be any string. */
  withGender() {
    query[0] = query[0].set({ subField: "gender", level })
    return this
  },
  /** The character's birth date */
  withDateOfBirth(fn: Fn<typeof FuzzyDate>) {
    query[0] = query[0].set({ subField: "dateOfBirth", level, hasSubField: true })
    let tmpQuery
    fn(FuzzyDate({ query: tmpQuery = [query[0]], level: level + 1 }))
    query[0] = tmpQuery[0]
    return this
  },
  /** The character's age. Note this is a string, not an int, it may contain further text and additional ages. */
  withAge() {
    query[0] = query[0].set({ subField: "age", level })
    return this
  },
  /** The characters blood type */
  withBloodType() {
    query[0] = query[0].set({ subField: "bloodType", level })
    return this
  },
  /** If the character is marked as favourite by the currently authenticated user */
  withIsFavourite() {
    query[0] = query[0].set({ subField: "isFavourite", level })
    return this
  },
  /** If the character is blocked from being added to favourites */
  withIsFavouriteBlocked() {
    query[0] = query[0].set({ subField: "isFavouriteBlocked", level })
    return this
  },
  /** The url for the character page on the AniList website */
  withSiteUrl() {
    query[0] = query[0].set({ subField: "siteUrl", level })
    return this
  },
  /** Media that includes the character */
  withMedia(args: CharacterMediaArgs | undefined, fn: Fn<typeof MediaConnection>) {
    query[0] = query[0].set({ subField: "media", level, variables: args, hasSubField: true })
    let tmpQuery
    fn(MediaConnection({ query: tmpQuery = [query[0]], level: level + 1 }))
    query[0] = tmpQuery[0]
    return this
  },
  /** The amount of user's who have favourited the character */
  withFavourites() {
    query[0] = query[0].set({ subField: "favourites", level })
    return this
  },
  /** Notes for site moderators */
  withModNotes() {
    query[0] = query[0].set({ subField: "modNotes", level })
    return this
  },
})
